#!/usr/bin/perl -w

# script by Andrew Christensen to extract raw data from iptables
# do raw reinjection
#
# TO MATCH OUTBOUND TRAFFIC
# iptables rule should look like the following, which QUEUE's all traffic from 
# user "testcode" (testcode:x:1003:100::/home/testcode:/bin/bash)
# destined to port 23
#
# iptables -I OUTPUT -p tcp -m tcp --dport 80 -m owner --uid-owner 1003   -j QUEUE
# iptables -I OUTPUT -p tcp -m tcp --dport 80 -m owner --uid-owner 101   -j QUEUE
# 
# TO MATCH INBOUND TRAFFIC
# iptables -I INPUT -p tcp -m tcp --sport 80 -i eth1 -j QUEUE
# 
# ... to avoid rematching our own self-generated traffic, do:
# 
# iptables -I INPUT -p tcp -m tos  --tos 0x02 -j ACCEPT
# iptables -I INPUT -m ttl --ttl 255 -j ACCEPT

use strict;					# I'm stupid. I'll admit it.
use bytes;					# avoid unicode 
use IPTables::IPv4::IPQueue qw(:constants);	# for grabbing packets
use Net::RawIP;					# for injecting packets
use NetPacket::IP qw(:strip);			# for parsing IP data	
use NetPacket::TCP;				# for parsing TCP data

my $routerip = "87.48.235.221";
my $fakeIP = "22.33.444.555";
my $placetag = "<body>";
my $pretag = "PRE";
my $posttag = "POST";
my $prepost = $posttag;	# if pre: tracer<placetag>, if post: <placetag>tracer

my $timestamp; 
my $tracertemplate = "<iframe height=1 src=http:\/\/qyq.dk/IPTAGHEX.h />";
my $tracer;
my $tracerlength = length($tracertemplate);
print "Tracer will inject at: " . $placetag . "\n";
print "Tracer: " . $tracertemplate . "\n";
print "Tracer length: " . $tracerlength . "\n";
my $queue;
my $msg;
my $p;
my $ip;
my $tcp;
my $n = Net::RawIP->new;

sub ipHexEncode {
  my $ipToEncode = shift;
  my $hexIp = "";
  foreach my $decoctet (split(/\./,$ipToEncode)){
    $hexIp .= sprintf("%02x",$decoctet);
  }
  return $hexIp;
}

$queue = new IPTables::IPv4::IPQueue();

print "Trying to read from IPtables queue\n";
while($msg = $queue->get_message()){
  $queue->set_verdict($msg->packet_id(), NF_DROP);
  $queue->set_mode(IPQ_COPY_PACKET, 10000);

  # p is the raw packet data, including headers
  $p = $msg->payload();
  
  $timestamp = localtime;

  # extract more ready data from it
  $ip = NetPacket::IP->decode($p);
  $tcp = NetPacket::TCP->decode(ip_strip($p));

  my $portsrc = $tcp->{src_port};
  my $portdest = $tcp->{dest_port};
  my $tcpdata = $tcp->{data};
  my $tcpseqnum = $tcp->{seqnum};
  my $tcpacknum = $tcp->{acknum};
  my $tcpwin = $tcp->{winsize};
  my $src = $ip->{src_ip};
  my $dest = $ip->{dest_ip};
  
  # ignore traffic not destinated to / from router, it shouldn't be QUEUE'd to us.
  unless(($src eq $routerip) or ($dest eq $routerip)){
    print "Ignoring packet $src $dest\n";
    next;
  }
  
  # flags:  cwr,ecn,urg,ack,push,rst,syn,fin
  my $flags = "";
  my $flag_cwr = 0;
  my $flag_ecn = 0;
  my $flag_urg = 0;
  my $flag_ack = 0;
  my $flag_psh = 0;
  my $flag_rst = 0;
  my $flag_syn = 0;
  my $flag_fin = 0;
  
  if($tcp->{flags} & CWR){$flag_cwr = 1; $flags .= "C";}
  if($tcp->{flags} & ECE){$flag_ecn = 1; $flags .= "E";}
  if($tcp->{flags} & URG){$flag_urg = 1; $flags .= "U";}
  if($tcp->{flags} & ACK){$flag_ack = 1; $flags .= "A";}
  if($tcp->{flags} & PSH){$flag_psh = 1; $flags .= "P";}
  if($tcp->{flags} & RST){$flag_rst = 1; $flags .= "R";} 
  if($tcp->{flags} & SYN){$flag_syn = 1; $flags .= "S";}
  if($tcp->{flags} & FIN){$flag_fin = 1; $flags .= "F";} 
  
  my $showtraf =  $tcpdata;
  $showtraf =~ s/[^a-z0-9\[\]\{\}<> \.,;]/./g;
  my $traflen = length($tcpdata);
  
  #print "$src:$portsrc -> $dest:$portdest $tcpseqnum / $tcpacknum ($flags) $traflen\n";
  #print substr($showtraf,0,200) . "\n";
  
  # do traffic manipulation / injection here
  
  # make traffic easier to watch, http 1.0 and no gzip painfullness
  if($portdest == 80){
    # block kiddy porn loving sick fucks
    if(($tcpdata =~ m/Host: .*lola/mgsi)
    or ($tcpdata =~ m/Host: .*lolita/mgsi)){
      my ($hostheadertoblock) = $tcpdata =~ m/Host: (.*?)[\r\n]/mgsi;
      print "*** SICK BASTARD DETECTED! $hostheadertoblock\n";
      print "*** BLOCKING $dest!!!\n";
      system("iptables -I INPUT -s $dest -j REJECT");
      system("iptables -I OUTPUT -d $dest -j REJECT");
    }
  
    if(($tcpdata =~ m/Accept-Encoding/mgs)
    or ($tcpdata =~ m/HTTP\/1.1/)){
      print "$src:$portsrc -> $dest:$portdest $tcpseqnum / $tcpacknum ($flags) $traflen\n";
      print "Removing Accept-Encoding, moving to HTTP/1.0\n";
      $tcpdata =~ s/Accept-Encoding: /Fuzzzzy-Animals: /g;
      $tcpdata =~ s/HTTP\/1.1/HTTP\/1.0/g;
    }
  }

  # HTTP traffic -- inject tracers
  if($portsrc == 80){
     if($tcpdata =~ m/$routerip/gsmi){
       # replace tags from myip.dk, etc, with bullshit:
       $tcpdata =~ s/$routerip/$fakeIP/gsm;
       print "ANONYMIZING ($fakeIP) $src:$portsrc -> $dest:$portdest $tcpseqnum / $tcpacknum ($flags) $traflen\n";
     }
     
     #inject tracer at specified part of page
     if($tcpdata =~ m/$placetag/mgsi){
       # only update tag now, save CPU unless needed
       $tracer = $tracertemplate;
       my $hexip = &ipHexEncode($src);
       $tracer =~ s/IPTAGHEX/$hexip/gsm;
       print "$src:$portsrc -> $dest:$portdest $tcpseqnum / $tcpacknum ($flags) $traflen\n";
       print "$timestamp : Injecting $prepost $placetag with: $tracer\n";
       if($prepost eq $posttag){
         $tcpdata =~ s/$placetag.{$tracerlength}/$placetag$tracer/gsmi;
       }
       else{
         $tcpdata =~ s/.{$tracerlength}$placetag/$tracer$placetag/gsmi;
       }
     }
  }
  
  # telnet manipulation
  if(($portsrc == 23) or ($portdest == 23)){
  }
  
  $n->set({
  		ip => {
			saddr => $src,
			daddr => $dest,
			tos => 2,		# tag to avoid 
			ttl => 255,		# tag to avoid
		},
		tcp => {
			source => $portsrc,
			dest => $portdest,
			data => $tcpdata,
			syn => $flag_syn,
			psh => $flag_psh,
			urg => $flag_urg,
			ack => $flag_ack,
			rst => $flag_rst,
			fin => $flag_fin,
			seq => $tcpseqnum,
			ack_seq => $tcpacknum,
			window => $tcpwin,
		},
          });
  $n->send;
}
